/*
 * Copyright (C) 2022, KylinSoft Co., Ltd.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

import QtQuick 2.12
import QtQuick.Controls 2.5
import QtQuick.Layouts 1.3
import "Controller.js" as Controller
import AppWindowFilterModel 1.0
import org.ukui.positioner 1.0

FocusScope {
    id : root
    property var appAreaController : new Controller.AppController(multitaskViewModel, root, multitaskViewManager, ["none"])
    property int currentTabIndex: 0

    AppWindowFilterModel {
        id: appAreaModel
        desktopId: rootWindow.highlightDesktop
        screenId: mainView.screenIndex
        sourceModel: appWindowListModel
    }

    Flickable {
        id : scrollArea
        anchors.fill : parent
        // 预览图布局区域 y 轴起始高度
        property int scrollViewYStartPoint : 32 * scaleFactor
        // 预览图布局区域距离左侧和右侧的间距
        property int scrollViewLeftSpace : isVerticalLayout ? 32 : 0
        anchors.leftMargin: scrollViewLeftSpace //注意：间距为0则无法显示滚动条
        anchors.rightMargin: scrollViewLeftSpace
        anchors.topMargin: 0
        anchors.bottomMargin: 0
        property ListModel windowPosInfoModel: ListModel{}
        property var rowSpacing : isVerticalLayout ? Math.floor(16 * scaleFactor) : Math.floor(10 * scaleFactor)
        property var columnSpacing : isVerticalLayout ? Math.floor(16 * scaleFactor) : Math.floor(10 * scaleFactor)
        property bool tabBorderVisble : false
        height: parent.height
        width: parent.width
        contentWidth : parent.width - scrollViewLeftSpace * 2
        contentHeight : Math.max(flow.height, height)
        clip : true

        ScrollBar.vertical: ScrollBar {
            id: vScrollBar
            parent: scrollArea.parent
            anchors.top: scrollArea.top
            anchors.left: scrollArea.right
            anchors.bottom: scrollArea.bottom
            visible : flow.height > scrollArea.height
        }

        MouseArea{
            anchors.fill : parent
            onClicked : {
                appAreaController.closeMultitaskView()
            }
        }


        NumberAnimation {
            id : scrollMoveAnimation
            target : vScrollBar
            property : "position"
            from : vScrollBar.position
            duration : 300
            easing.type : Easing.InOutQuart
        }

        UkuiFlow {
            id : flow
            property bool needFlowYAnimation : false

            spacing : scrollArea.rowSpacing
            topPadding: scrollArea.scrollViewYStartPoint
            width : parent.width
            flow : Flow.LeftToRight
            layoutDirection: isLayoutLeftToRight ? Qt.LeftToRight : Qt.RightToLeft
            lineHorizontalCenterAlignment: true

            move :Transition { //子项增加时或者减少时的过度，也就是子项的visible变化的过度
                NumberAnimation {
                    properties : "x,y";
                    easing.type : Easing.InOutQuad;
                    duration : isVerticalLayout ? 500 : 300
                }
            }

            onHeightChanged: {
                y = Math.max((scrollArea.height - flow.height) / 2, 0)
            }

            // flow 垂直居中动画
            Behavior on y {
                enabled: flow.needFlowYAnimation
                NumberAnimation {
                    duration : isVerticalLayout ? 500 : 300
                    easing.type : Easing.InOutQuart
                }
            }

            onPositioningComplete: {
                updateWindowPosInfo()
            }

            function updateWindowPosInfo() {
                var rowWindowNum = 0  //窗口区域每行的窗口数
                var leftX = 0
                var rightX = 0
                var rowY = flow.topPadding
                var itemCount = 0
                var winModel = []
                scrollArea.windowPosInfoModel.clear()

                for (let i = 0; i < flow.children.length; ++i) {
                    if (flow.children[i].width === 0) {
                        continue
                    }

                    // 根据预览窗 Y 坐标判断是否到下一行
                    if (flow.children[i].y !== rowY) {
                        if (winModel.length > 0) {
                            scrollArea.windowPosInfoModel.append({model : winModel, rowWindowCount : rowWindowNum})
                            winModel = []
                            rowWindowNum = 0
                        }
                        rowY += flow.children[i].height + flow.spacing
                    }

                    // 更新预览窗在 Flow 布局中的 X 坐标并存入局部 model
                    itemCount++
                    leftX = flow.children[i].x
                    rightX = leftX + flow.children[i].width
                    winModel.push({index: itemCount - 1, leftX: leftX, rightX: rightX})
                    rowWindowNum++
                }
                if (winModel.length > 0) {
                    scrollArea.windowPosInfoModel.append({model : winModel, rowWindowCount : rowWindowNum})
                }
            }

            Repeater {
                id : previewWindowRepeater
                model : appAreaModel
                activeFocusOnTab: count !== 0 ? true : false
                focus : true

                delegate: Rectangle {
                    id : thumbnailitem
                    width: appWindow.width
                    height: appWindow.height
                    focus: true
                    color : "transparent"
                    property var thumbWidth : getThumbnailSize()[0]
                    property var thumbHeight : getThumbnailSize()[1]
                    property string uuid: model.Uuid

                    states: [
                        State {
                            name : "tabActive"
                            when: previewWindowRepeater.activeFocus && index === currentTabIndex && (multitaskViewManager.shouldShowTabBorder() || scrollArea.tabBorderVisble)
                            PropertyChanges {
                                target: appWindow.borderRectangle
                                borderColor : tabFocusBorderColor
                            }
                        }
                    ]

                    function getThumbnailSize()
                    {
                        var thumbHeight = thumbRectangleHeight
                        var max_width = parent.width / 3
                        var clientWidth = model.Width
                        var clientHeight = model.Height
                        var thumbWidth = Math.floor(thumbHeight * clientWidth / clientHeight)
                        if (thumbWidth > max_width) {
                            thumbWidth = max_width
                            thumbHeight = Math.floor(thumbWidth * clientHeight / clientWidth)
                        }
                        return [thumbWidth, thumbHeight]
                    }

                    AppPreviewWindow {
                        id : appWindow
                        // Warning: 因为 appWindow 的 parent 会改变，所以不要使用 parent 获取宽高
                        minWidth: Math.floor(root.width / 10)

                        thumbHeight : thumbnailitem.thumbHeight
                        thumbWidth : thumbnailitem.thumbWidth
                        windowType : "app"
                        x_wid : model.Id
                        wid : model.Uuid
                        caption : model.Caption
                        parent: thumbnailitem

                        Drag.onActiveChanged : {
                            if (!Drag.active) {
                                appAreaController.onSourceDropped(appWindow, rootWindow.highlightDesktop, rootWindow.highlightDesktop)
                            } else {
                                //拖拽动画开始前,修改 appWindow 父对象
                                appWindow.parent = mainView
                                zoomOutAnimation()
                            }
                        }
                    }
                }//Rectangle

                onActiveFocusChanged: {
                    if (previewWindowRepeater.activeFocus) {
                        currentTabIndex = 0
                        if (vScrollBar.position > 0) {
                            scrollMoveAnimation.to = 0
                            scrollMoveAnimation.start()
                        }
                        multitaskViewManager.activeAreaChangeDataCollect("Windows Preview Area")
                    }
                }

                onItemAdded : {
                    if (flow.needFlowYAnimation !== false) {
                        flow.needFlowYAnimation = false
                    }
                }

                onItemRemoved : {
                    if (flow.needFlowYAnimation !== true) {
                        flow.needFlowYAnimation = true
                    }
                }

                Keys.onPressed: {
                    if (event.key === Qt.Key_Left) {
                        scrollArea.tabBorderVisble = true
                        if (currentTabIndex > 0) {
                            currentTabIndex = currentTabIndex - 1
                        } else if (currentTabIndex === 0) {
                            currentTabIndex = previewWindowRepeater.count - 1
                        }
                        event.accepted = true

                    } else if (event.key === Qt.Key_Right ) {
                        scrollArea.tabBorderVisble = true
                        if (currentTabIndex < previewWindowRepeater.count - 1) {
                            currentTabIndex = currentTabIndex + 1
                        } else if (currentTabIndex === previewWindowRepeater.count - 1) {
                            currentTabIndex = 0
                        }
                        event.accepted = true

                    } else if (event.key === Qt.Key_Up || event.key === Qt.Key_Down) {
                        scrollArea.tabBorderVisble = true
                        var isKeyUp = event.key === Qt.Key_Up
                        currentTabIndex = appAreaController.keyUpDownSelectWindow(isKeyUp, scrollArea.windowPosInfoModel, currentTabIndex)
                        event.accepted = true

                    } else if (event.key === Qt.Key_Tab) {
                        if (event.modifiers & Qt.MetaModifier) {
                            appAreaController.closeMultitaskView()
                            return
                        }
                        scrollArea.tabBorderVisble = true
                        currentTabIndex = 0
                        event.accepted = false

                    } else if (event.key === Qt.Key_Return || event.key === Qt.Key_Enter || event.key === Qt.Key_Space) {
                        appAreaController.onWindowClicked(previewWindowRepeater.itemAt(currentTabIndex).uuid);
                        event.accepted = true

                    } else if (event.key === Qt.Key_End) {
                        scrollArea.tabBorderVisble = true
                        currentTabIndex = previewWindowRepeater.count - 1
                        event.accepted = true

                    } else if (event.key === Qt.Key_Home) {
                        scrollArea.tabBorderVisble = true
                        currentTabIndex = 0
                        event.accepted = true

                    } else if (event.key === Qt.Key_PageUp || event.key === Qt.Key_PageDown) {
                        scrollArea.tabBorderVisble = true
                        var isKeyPageUp = event.key === Qt.Key_PageUp
                        currentTabIndex = appAreaController.keyPageUpDownSelectWindow(isKeyPageUp, scrollArea.windowPosInfoModel, currentTabIndex)
                        event.accepted = true

                    } else if (event.key === Qt.Key_Escape) {
                        appAreaController.closeMultitaskView()
                    }

                    if (event.accepted) {
                        scrollToTabItem()
                    }
                }

                function scrollToTabItem() {
                    var currentNextY = itemAt(currentTabIndex).mapToItem(scrollArea, 0, 0).y
                    var previewWindowHeight = itemAt(currentTabIndex).height
                    if (currentNextY < 0) {
                        scrollMoveAnimation.to = vScrollBar.position - (-currentNextY) / scrollArea.contentHeight
                        scrollMoveAnimation.start()
                    } else if (currentNextY + previewWindowHeight > scrollArea.height) {
                        scrollMoveAnimation.to = vScrollBar.position + (currentNextY + previewWindowHeight - scrollArea.height) / scrollArea.contentHeight
                        scrollMoveAnimation.start()
                    }
                }
            }//Repeater
            Component.onCompleted: {
                previewWindowRepeater.forceActiveFocus()
            }
        }//Flow
    }//ScrollView
}
